@@ -2,15 +2,12 @@ from datetime import datetime
 from functools import cached_property
 from json import dumps
 from json import loads
-from os import listdir
+from os import PathLike
 from os import makedirs
-from os.path import abspath
-from os.path import dirname
-from os.path import isdir
-from os.path import isfile
-from os.path import join
+from pathlib import Path
 from re import sub
 from shutil import copy
+from shutil import move
 from sqlite3 import Connection
 from sqlite3 import Cursor
 from sqlite3 import DatabaseError
@@ -55,21 +52,40 @@ def guess_extension(file: Optional[bytes], default: str = "") -> str:
         return ext
 
 
-def tiered_path(id_: Union[int, str], depth: int = 5, width: int = 2) -> str:
+def tiered_path(id_: Union[int, str], depth: int = 5, width: int = 2) -> Path:
     assert isinstance(id_, int) or (isinstance(id_, str) and id_.isdigit()), "id not an integer"
     assert isinstance(depth, int) and depth > 0, "depth lower than 0"
     assert isinstance(width, int) and width > 0, "depth lower than 0"
 
     id_str: str = str(int(id_)).zfill(depth * width)
-    return join(*[id_str[n:n + width] for n in range(0, depth * width, width)])
+    return Path(*[id_str[n:n + width] for n in range(0, depth * width, width)])
 
 
-def copy_folder(src: str, dest: str):
-    if isdir(src):
-        for item in listdir(src):
-            copy_folder(join(src, item), join(dest, item))
-    elif isfile(src) and not isfile(dest):
-        makedirs(dirname(dest), exist_ok=True)
+def check_version(version_a: str, raise_for_error: bool = True, *, major: bool = True, minor: bool = True,
+                  patch: bool = True, version_b: str = __version__
+                  ) -> Optional[DatabaseError]:
+    err: Optional[DatabaseError] = None
+    if version_a == version_b:
+        return None
+    elif not version_a:
+        err = DatabaseError(f"version error: {version_a}")
+    elif (v_a := version_a.split("."))[0] != (v_b := version_b.split("."))[0]:
+        err = DatabaseError(f"major version is not latest: {v_a[0]} != {v_b[0]}") if major else None
+    elif v_a[1] != v_b[1]:
+        err = DatabaseError(f"minor version is not latest: {v_a[1]} != {v_b[1]}") if minor else None
+    elif v_a[2] != v_b[2]:
+        err = DatabaseError(f"patch version is not latest: {v_a[2]} != {v_b[2]}") if patch else None
+    if err is not None and raise_for_error:
+        raise err
+    return err
+
+
+def copy_folder(src: Path, dest: Path):
+    if src.is_dir():
+        for item in src.iterdir():
+            copy_folder(item, dest / item.name)
+    elif src.is_file and not dest.is_file():
+        makedirs(dest.parent, exist_ok=True)
         copy(src, dest)
 
 
@@ -77,27 +93,29 @@ def clean_username(username: str) -> str:
     return str(sub(r"[^a-zA-Z0-9\-.~,]", "", username.lower().strip()))
 
 
-def copy_cursors(db_a: 'FADatabase', db_b: 'FADatabase', *cursors: 'FADatabaseCursor', replace: bool = True):
-    """
-    A -> B
-    """
-    assert all(c.table.database.database_path == db_a.database_path for c in cursors), \
-        "Cursors must point to the database calling copy"
+def copy_cursors(db_dest: 'FADatabase', cursors: list['FADatabaseCursor'] = None, replace: bool = True):
+    if not cursors:
+        return
+    assert (e := check_version(version_dest := db_dest.version, raise_for_error=False, patch=False)) is None, \
+        f"Destination database is not up to date. {e.args[0]}"
+    assert all(
+        check_version(c.table.database.version, raise_for_error=False, patch=False, version_b=version_dest) is None
+        for c in cursors), "Cursors must point to a database with the same version as the destination database"
     assert all(set(c.columns) == set(c.table.columns) for c in cursors), "Cursors must contain all table columns"
-    db_a.check_version(patch=False, version=db_b.version)
-    db_a.check_version(patch=False)
 
-    db_a_files_folder: str = join(dirname(db_a.database_path), db_a.settings["FILESFOLDER"])
-    db_b_files_folder: str = join(dirname(db_b.database_path), db_b.settings["FILESFOLDER"])
+    dest_files: Path = db_dest.files_folder
+    cursor_files: Optional[Path] = None
 
     for cursor in cursors:
-        table_b: FADatabaseTable = db_b[cursor.table.table]
+        table_dest: FADatabaseTable = db_dest[cursor.table.table]
+        if cursor.table.table == submissions_table:
+            cursor_files = cursor.table.database.files_folder
         for entry in cursor:
-            if not replace and table_b.table == users_table and (entry_b := table_b[entry[table_b.column_id]]):
+            if not replace and table_dest.table == users_table and (entry_b := table_dest[entry[table_dest.column_id]]):
                 entry["FOLDERS"] = list(set(entry["FOLDERS"] + entry_b["FOLDERS"]))
-            if table_b.table == submissions_table:
-                copy_folder(join(db_a_files_folder, p := tiered_path(entry["ID"])), join(db_b_files_folder, p))
-            table_b.insert(cursor.table.format_dict(entry), replace=True)
+            if table_dest.table == submissions_table:
+                copy_folder(cursor_files / (p := tiered_path(entry["ID"])), dest_files / p)
+            table_dest.insert(cursor.table.format_dict(entry), replace=True)
 
 
 def format_list(obj: list[Value]) -> str:
@@ -277,17 +295,23 @@ class FADatabaseSettings(FADatabaseTable):
 
 
 class FADatabaseSubmissions(FADatabaseTable):
+    def get_submission_files(self, submission_id: int) -> tuple[Optional[Path], Optional[Path]]:
+        if (entry := self[submission_id]) is None or (f := entry["FILESAVED"]) == 0:
+            return None, None
+        folder: Path = self.database.files_folder / tiered_path(submission_id)
+        file: Path = folder / ("submission" + f".{(ext := entry['FILEEXT'])}" * bool(ext)) if f >= 10 else None
+        thumb: Path = folder / "thumbnail.jpg" if f % 10 == 1 else None
+        return file, thumb
+
     def save_submission_file(self, submission_id: int, file: Optional[bytes], name: str, ext: str,
                              guess_ext: bool = True) -> str:
         if file is None:
             return ""
 
         ext = guess_extension(file, ext) if guess_ext else ext
-        path: str = join(dirname(self.database.database_path), self.database.settings["FILESFOLDER"],
-                         tiered_path(submission_id), name + f".{ext}" * bool(ext))
-
-        makedirs(dirname(path), exist_ok=True)
-        open(path, "wb").write(file)
+        folder: Path = self.database.files_folder / tiered_path(submission_id)
+        makedirs(folder, exist_ok=True)
+        open(folder / (name + f".{ext}" * bool(ext)), "wb").write(file)
 
         return ext
 
@@ -357,17 +381,22 @@ class FADatabaseCursor:
         self.table: FADatabaseTable = table
         self.cursor: Cursor = cursor
         self.columns: list[str] = columns
+        self.entries: Generator[Entry, None, None] = (
+            {k: unpack_list(v) if k in self.table.list_columns else v for k, v in zip(columns, entry)}
+            for entry in self.cursor
+        )
+
+    def __next__(self) -> Entry:
+        return next(self.entries)
 
     def __iter__(self) -> Generator[Entry, None, None]:
-        columns = list(map(str.upper, self.columns))
-        return ({k: unpack_list(v) if k in self.table.list_columns else v for k, v in zip(columns, entry)}
-                for entry in self.cursor)
+        return self.entries
 
 
 class FADatabase:
-    def __init__(self, database_path: str):
-        self.database_path: str = abspath(database_path)
-        self.connection: Connection = connect(database_path)
+    def __init__(self, database_path: Union[str, PathLike]):
+        self.database_path: Path = Path(database_path).resolve()
+        self.connection: Connection = connect(self.database_path)
 
         if journals_table not in (tables := self.tables):
             make_journals_table(self.connection)
@@ -423,6 +452,17 @@ class FADatabase:
     def is_clean(self) -> bool:
         return self.total_changes == self.committed_changes
 
+    @property
+    def files_folder(self) -> Path:
+        return self.database_path.parent / self.settings["FILESFOLDER"]
+
+    def move_files_folder(self, new_folder: Union[str, PathLike]) -> Path:
+        new_folder = Path(new_folder)
+        new_folder = new_folder if new_folder.is_absolute() else (self.database_path.parent / new_folder)
+        move(self.files_folder, new_folder)
+        self.settings["FILESFOLDER"] = str(new_folder.relative_to(self.database_path.parent))
+        return new_folder
+
     def upgrade(self):
         self.connection = update_database(self.connection, __version__)
 
@@ -435,20 +475,8 @@ class FADatabase:
     def check_version(self, raise_for_error: bool = True, *, major: bool = True, minor: bool = True, patch: bool = True,
                       version: str = __version__
                       ) -> Optional[DatabaseError]:
-        err: Optional[DatabaseError] = None
-        if (v := self.version) == version:
-            return None
-        if not v:
-            err = DatabaseError(f"version error: {v}")
-        if (v_db := v.split("."))[0] != (v_md := version.split("."))[0]:
-            err = DatabaseError(f"major version is not latest: {v_db[0]} != {v_md[0]}") if major else None
-        elif v_db[1] != v_md[1]:
-            err = DatabaseError(f"minor version is not latest: {v_db[1]} != {v_md[1]}") if minor else None
-        elif v_db[2] != v_md[2]:
-            err = DatabaseError(f"patch version is not latest: {v_db[2]} != {v_md[2]}") if patch else None
-        if err is not None and raise_for_error:
-            raise err
-        return err
+        return check_version(self.version, raise_for_error=raise_for_error, major=major, minor=minor, patch=patch,
+                             version_b=version)
 
     def check_connection(self, raise_for_error: bool = True) -> list[Process]:
         ps: list[Process] = []
@@ -477,11 +505,10 @@ class FADatabase:
         self.connection.close()
 
     def merge(self, db_b: 'FADatabase', *cursors: FADatabaseCursor):
-        copy_cursors(db_b, self, *cursors or (db_b.users.select(), db_b.submissions.select(), db_b.journals.select()),
-                     replace=False)
+        copy_cursors(self, cursors or [db_b.users.select(), db_b.submissions.select(), db_b.journals.select()], False)
 
     def copy(self, db_b: 'FADatabase', *cursors: FADatabaseCursor):
-        copy_cursors(self, db_b, *cursors or (self.users.select(), self.submissions.select(), self.journals.select()))
+        copy_cursors(db_b, cursors or [self.users.select(), self.submissions.select(), self.journals.select()])
 
     def vacuum(self):
         self.connection.execute("VACUUM")
